/*
Library for 16-channel LED Driver TLC5922, TLC5923, TLC5924 
This Code is for SPI control using an unlimited number of drivers connected daisy-chained

Author: Kai Franke
Date: 02.02.2008

This Code is free for all, do whatever you want. Improvements/comments/wishes are welcome: kaifrankes "at" web.de  or ICQ: 177183299


Please change the settings below for your specific controller (default: ATmega8)
Check the datasheet for details    

Compile this code with CodevisionAVR (compile for speed and the trial version will be enough)

Connect the drivers with a C as follows:

Controller      Driver1         Driver2         Driver3
--------------------------------------------------------  
SCK             SCLK            SCLK            SCLK
SS              XLAT            XLAT            XLAT
PORTB.0         BLANK           BLANK           BLANK
PORTB.1         MODE            MODE            MODE
MOSI            SIN
                SOUT            SIN
                                SOUT            SIN
                IREF --> resistor (min. 600 Ohm) to GND for each driver

the most important funktions are:

initialize()                    // initializes SPI and Timer1 (change if not using ATmega8)
reset();                        // reset all arrays to 0 and set all LEDs to ON, dot correction 0 
set_rgb_on()                    // transmits the ON/OFF settings for RGB LEDs saved in rgb_on[]
color_flow_on(UINT speed)       // smooth color switching on, speed is an int value to set changing speed
color_flow_off()                // smooth color switching off
running_light_on(UINT speed)    // running light on, speed is an int value to set running speed
running_light_off()             // running light off

// sets all RGB LEDs the same color
set_global_color(UCHAR value_red, UCHAR value_green, UCHAR value_blue)

*/


// Number of drivers connected
#define drivers         2

// Settings for SPI
#define SPI_PORT        PORTB  
#define SPI_PORT_DIR    DDRB
#define MISO            4
#define MOSI            3
#define SCK             5
#define CS              2   
#define BLANK           0
#define MODE            1

#define MSTR            4
#define SPR0            0
#define SPR1            1
#define SPE             6
#define SPI2X           0
#define SPIF            7

// you might need to adjust the Timer settings, too if not using ATmega8 !!!

// no need to change anything else if using Codevision and ATmega8
#define CS_ON()         SPI_PORT.CS = 0;
#define CS_OFF()        SPI_PORT.CS = 1;

#define DOT_CORRECTION  SPI_PORT.MODE = 1;
#define ON_OFF          SPI_PORT.MODE = 0;
#define ALL_ON()        SPI_PORT.BLANK = 0;
#define ALL_OFF()       SPI_PORT.BLANK = 1;   

#define UCHAR unsigned char   
#define UINT  unsigned int

// Variables decleration
                                  
UINT on[drivers];                       // stores the ON/OFF settings of each single output
UINT rgb_on[drivers];                   // stores the ON/OFF settings of each RGB LED (5 RGBs/driver)
UINT red;                               // value between 0..127 is valid, where 0 is OFF and 127 ist full
UINT green;                             // value between 0..127 is valid, where 0 is OFF and 127 ist full
UINT blue;                              // value between 0..127 is valid, where 0 is OFF and 127 ist full
UCHAR dot_correction[14][drivers];      // stores the brightness settings in a format to transmit them via SPI
UCHAR single_color[16][drivers];        // stores the brightness settings with one char for 7 Bit (MSB is a don't care)
UINT puffer;
UINT position = 0;
UINT i;
int j;
UINT led1;
UINT led2;
UINT led3;
UINT led4;
UINT led5;
    
// inline HW SPI routine for best speed performance
inline void send_byte(UCHAR Byte)
{
    SPDR = Byte;
    while(!(SPSR & (1<<SPIF)));
}   

// refresh all driver's dot correction values (dot correction = brightness of each Color)
// reads the array dot_correction[][]
void refresh_dot_correction()
{
        DOT_CORRECTION
        CS_ON()  
        for(j = (drivers-1);j>=0;j--)
        {
                for(i = 0;i<14;i++)
                {
                        send_byte(dot_correction[13-i][0]);  
                }      
        }
        CS_OFF()
}

// refresh all driver's ON/OFF settings
// reads the array on[]
void refresh_on()
{
        ON_OFF          // set MODE to send ON/OFF settings
        CS_ON()
        for(j = (drivers-1);j>=0;j--)
        {
                puffer = on[j]>>8;
                send_byte(puffer);
                send_byte(on[j]);        
        }              
        //ALL_OFF()       // refering to the datasheet, this is needed
        CS_OFF() 
        //ALL_ON()        // practical tests showed the opposite
}   

// reset all arrays to 0 and set all LEDs to ON, dot correction 0
void reset()
{        
        // set brightness of LEDs to 0      
        for(j = 0;j<drivers;j++)
        {
                for(i = 0;i<14;i++)
                {
                        dot_correction[i][j] = 0; 
                }  
        }
        
        // set color array to 0
        for(j = 0;j<drivers;j++)
        {
                for(i = 0;i<16;i++)
                {
                        single_color[i][j] = 0; 
                }      
        } 
        refresh_dot_correction();       // update brightness (All brightness 0)   
        
        // set all LEDs to ON
        for(j = 0;j<drivers;j++)
        {
                on[j] = 0xFFFF;  
        }  
        refresh_on();                   // update ON/OFF settings (ALL ON)
}

// initializes SPI and Timer1 (change if not using ATmega8)
void initialize()
{
        // Set SPI Pins   
        SPI_PORT_DIR.MOSI = 1;
        SPI_PORT_DIR.MISO = 0;
        SPI_PORT_DIR.SCK = 1;
        SPI_PORT_DIR.CS = 1; 
        
        // Set BLANK and MODE Pins
        SPI_PORT_DIR.BLANK = 1;
        SPI_PORT_DIR.MODE = 1;
        
        // initialize SPI PORT
        // Master, full Speed, Mode 0, MSB first
        SPCR = (1<<SPE)|(1<<MSTR)|(0<<SPR0)|(0<<SPR1);
        SPSR = (1<<SPI2X); 
                    
        // initialize all LEDs ON
        reset();
        
        // initialize Timer and enable Interrupts
        TCCR1B=0x05; 
        TIMSK=0;
        #asm("sei") 
}  

// converts the 7 Bit color value saved in single_color[][] to a format for SPI transmission and saves it in dot_correction[][]
void Array_Conversion()
{
        for(j = 0;j<drivers;j++)
        {
                for(i = 0;i<7;i++)
                {
                         
                        dot_correction[i][j] = (single_color[i][j] >> (i));                        
                        dot_correction[i][j] |= (single_color[i+1][j] << (7-i));
                }
                for(i = 0;i<7;i++)
                {
                        dot_correction[i+7][j] =  (single_color[i+8][j] >> (i));                       
                        dot_correction[i+7][j] |= (single_color[i+9][j] << (7-i));
                }
        } 
} 

// Timer1 CompareA ISR
// needed for the color_flow effect
interrupt [TIM1_COMPA] void timer1_compa_isr(void)
{
        TCNT1 = 0;
        
        if (red == 127 && green < 127 && blue == 0)
        {
                green++;
        }
        
        if (red > 0 && green == 127 && blue == 0)
        {
                red--;
        }
        
        if (red == 0 && green == 127 && blue < 127)
        {
                blue++;
        }
        
        if (red == 0 && green > 0 && blue == 127)
        {
                green--;
        }
        
        if (green == 0 && blue == 127 && red < 127)
        {
                red++;
        }
        
        if (green == 0 && blue > 0 && red == 127)
        {
                blue--;
        }
        
        for(j = 0;j<drivers;j++)
        {
                for(i = 0;i<15;i=i+3)
                {
                        single_color[i][j] = red;
                        single_color[i+1][j] = green;
                        single_color[i+2][j] = blue;
                }
                single_color[15][j] = 0;
        }
        Array_Conversion();   
        refresh_dot_correction(); 
        

}

// Timer1 CompareB ISR    
// needed for the running light
interrupt [TIM1_COMPB] void timer1_compb_isr(void)
{
        TCNT1 = 0;
        refresh_on();
        if (on[position] >= 0b0010000000000000)
        {
                if ((position + 1) < drivers)
                {
                        position++; 
                        on[position] = on[position-1]>>12;
                        on[position-1] = 0;
                }
                else
                {
                        position = 0;
                        on[0] = 0b0000000000000111;
                }
        }
        else
        {
                on[position] = on[position]<<3;
        }                                   
}

// smooth color switching on, speed is an int value to set changing speed
void color_flow_on(UINT speed)
{             
        OCR1A = speed;          // set compare match
        red = 127;              // start with pure red
        green = 0;
        blue = 0; 
        TIMSK = 0x10;           // enable Interrupt
        TCNT1 = 0;              // set Timer to 0
}

// smooth color switching off
void color_flow_off()
{
        TIMSK = 0;              // disable Interrupts
}
                   
// running light on, speed is an int value to set running speed
void running_light_on(UINT speed)
{
        // set all LEDs to OFF
        for(j = 0;j<drivers;j++)
        {
                on[j] = 0;  
        }  
        OCR1B = speed;          // set compare match
        TIMSK = 0x08;           // enable Interrupt
        TCNT1 = 0;              // set Timer to 0
        on[0] = 0b0000000000000111;     // start with LED 0 driver 0 ON
           
}   

// running light off
void running_light_off()
{
        TIMSK = 0;              // disable Interrupt
}              

// sets all RGB LEDs the same color
void set_global_color(UCHAR value_red, UCHAR value_green, UCHAR value_blue)
{
        for(j = 0;j<drivers;j++)
        {
                for(i = 0;i<15;i = i+3)
                {
                        single_color[i][0] = value_red; 
                        single_color[i+1][0] = value_green;
                        single_color[i+2][0] = value_blue;
                }
        } 
        Array_Conversion();
        refresh_dot_correction();
}  

// transmits the ON/OFF settings for RGB LEDs saved in rgb_on[]
void set_rgb_on()
{
        for(j = 0;j<drivers;j++)
        {
                led1 = (rgb_on[j] & 0b00010000) << 8;           // needed for Array Conversion
                led2 = (rgb_on[j] & 0b00001000) << 6;
                led3 = (rgb_on[j] & 0b00000100) << 4;
                led4 = (rgb_on[j] & 0b00000010) << 2;
                led5 = (rgb_on[j] & 0b00000001);
                on[j] = (led1 | (led1 << 1) | (led1 << 2) | led2 | (led2 << 1) | (led2 << 2) | led3 | (led3 << 1) | (led3 << 2) | led4 | (led4 << 1) | (led4 << 2) | led5 | (led5 << 1) | (led5 << 2));
        } 
        
        refresh_on();
}